import { getDocument } from "pdfjs-dist/legacy/build/pdf.mjs";
import * as opentype from "opentype.js";

const stubConsoles = (cb, outputStream = ["log", "error", "warn"]) => {
  let id = null;
  const oldLogs = outputStream.map((type) => console[type]);

  outputStream.forEach((type) => {
    console[type] = (...str) => {
      if (id === null) {
        cb(...str);
      } else {
        cb(id, ...str);
      }
    };
  });
  return {
    unStub: () => {
      outputStream.forEach((type, index) => {
        console[type] = oldLogs[index];
      });
    },
    setId: (newId) => {
      id = newId;
    },
  };
};

/**
 * space glyph is often missing when extracting embedded fonts from pdf.js.
 * since the glyph for space is a blank path, we can easily backfill it here.
 */
const extendFontWithSpaceGlyph = (fontBuffer) => {
  let originalFont;
  try {
    originalFont = opentype.parse(fontBuffer.buffer);
  } catch (err) {
    console.error("Error loading font:", err);
    return;
  }

  const spaceGlyph = new opentype.Glyph({
    name: "space",
    unicode: 32, // Unicode for space (U+0020)
    advanceWidth: 500,
    path: new opentype.Path(),
  });

  // Add the new glyph to the rebuilt font
  originalFont.glyphs.push(originalFont.glyphs.length, () => spaceGlyph);

  const { unStub } = stubConsoles(() => {});
  const buffer = originalFont.toArrayBuffer();
  unStub();
  return buffer;
};

const CMAP_URL = "../../node_modules/pdfjs-dist/cmaps/";
const STANDARD_FONT_DATA_URL = "../../node_modules/pdfjs-dist/standard_fonts/";
const CMAP_PACKED = false;

/**
 * @param {Object} obj
 * @param {string | URL | undefined} obj.url Local or http url
 * @param {string | Uint8Array | ArrayBuffer | undefined} obj.data Binary PDF data. Use TypedArrays (Uint8Array) to improve the memory usage. If PDF data is BASE64-encoded, use atob() to convert it to a binary string first.
 * @returns {Promise<Record<string, ArrayBuffer>>}
 */
async function extractFonts({ url, data }) {
  const initParam = {
    cMapUrl: CMAP_URL,
    cMapPacked: CMAP_PACKED,
    standardFontDataUrl: STANDARD_FONT_DATA_URL,
    fontExtraProperties: true,
    useSystemFonts: true,
  };
  if (url) {
    initParam.url = url;
  } else if (data) {
    initParam.data = data;
  } else {
    throw new Error("Either 'url' or 'data' must be provided");
  }

  const loadingTask = getDocument(initParam);
  const pdfDocument = await loadingTask.promise;
  const fontMap = {};

  for (let i = 1; i <= pdfDocument.numPages; i++) {
    const page = await pdfDocument.getPage(i);
    await page.getOperatorList();

    const text = await page.getTextContent();

    text.items.forEach((item) => {
      if (!fontMap[item.fontName]) {
        const font = page.commonObjs.get(item.fontName);
        fontMap[item.fontName] = extendFontWithSpaceGlyph(font.data);
      }
    });
    page.cleanup();
  }

  return fontMap;
}

export default extractFonts;
